@@ -17,6 +17,7 @@
 #include "configfile.h"
 #include "theme.h"
 #include "utility.h"
+#include "asserts.h"
 
 #include "creds/abstractcredentials.h"
 
@@ -67,6 +68,7 @@ static const char downloadLimitC[]    = "BWLimit/downloadLimit";
 
 static const char newBigFolderSizeLimitC[] = "newBigFolderSizeLimit";
 static const char useNewBigFolderSizeLimitC[] = "useNewBigFolderSizeLimit";
+static const char confirmExternalStorageC[] = "confirmExternalStorage";
 
 static const char maxLogLinesC[] = "Logging/maxLogLines";
 
@@ -138,7 +140,7 @@ void ConfigFile::setOptionalDesktopNotifications(bool show)
 void ConfigFile::saveGeometry(QWidget *w)
 {
 #ifndef TOKEN_AUTH_ONLY
-    Q_ASSERT(!w->objectName().isNull());
+    ASSERT(!w->objectName().isNull());
     QSettings settings(configFile(), QSettings::IniFormat);
     settings.beginGroup(w->objectName());
     settings.setValue(QLatin1String(geometryC), w->saveGeometry());
@@ -157,7 +159,7 @@ void ConfigFile::saveGeometryHeader(QHeaderView *header)
 {
 #ifndef TOKEN_AUTH_ONLY
     if(!header) return;
-    Q_ASSERT(!header->objectName().isEmpty());
+    ASSERT(!header->objectName().isEmpty());
 
     QSettings settings(configFile(), QSettings::IniFormat);
     settings.beginGroup(header->objectName());
@@ -170,7 +172,7 @@ void ConfigFile::restoreGeometryHeader(QHeaderView *header)
 {
 #ifndef TOKEN_AUTH_ONLY
     if(!header) return;
-    Q_ASSERT(!header->objectName().isNull());
+    ASSERT(!header->objectName().isNull());
 
     QSettings settings(configFile(), QSettings::IniFormat);
     settings.beginGroup(header->objectName());
@@ -183,14 +185,14 @@ QVariant ConfigFile::getPolicySetting(const QString &setting, const QVariant& de
     if (Utility::isWindows()) {
         // check for policies first and return immediately if a value is found.
         QSettings userPolicy(QString::fromLatin1("HKEY_CURRENT_USER\\Software\\Policies\\%1\\%2")
-                             .arg(APPLICATION_VENDOR).arg(Theme::instance()->appName()),
+                             .arg(APPLICATION_VENDOR, Theme::instance()->appName()),
                              QSettings::NativeFormat);
         if(userPolicy.contains(setting)) {
             return userPolicy.value(setting);
         }
 
         QSettings machinePolicy(QString::fromLatin1("HKEY_LOCAL_MACHINE\\Software\\Policies\\%1\\%2")
-                                .arg(APPLICATION_VENDOR).arg(APPLICATION_NAME),
+                                .arg(APPLICATION_VENDOR, APPLICATION_NAME),
                                 QSettings::NativeFormat);
         if(machinePolicy.contains(setting)) {
             return machinePolicy.value(setting);
@@ -229,8 +231,8 @@ QString ConfigFile::excludeFile(Scope scope) const
     // directories.
     QFileInfo fi;
 
-    if (scope != SystemScope) {
-        QFileInfo fi;
+    switch (scope) {
+    case UserScope:
         fi.setFile( configPath(), exclFile );
 
         if( ! fi.isReadable() ) {
@@ -240,12 +242,12 @@ QString ConfigFile::excludeFile(Scope scope) const
             fi.setFile( configPath(), exclFile );
         }
         return fi.absoluteFilePath();
-    } else if (scope != UserScope) {
+    case SystemScope:
         return ConfigFile::excludeFileFromSystem();
-    } else {
-        Q_ASSERT(false);
-        return QString(); // unreachable
     }
+
+    ASSERT(false);
+    return QString();
 }
 
 QString ConfigFile::excludeFileFromSystem()
@@ -255,7 +257,7 @@ QString ConfigFile::excludeFileFromSystem()
     fi.setFile( QCoreApplication::applicationDirPath(), exclFile );
 #endif
 #ifdef Q_OS_UNIX
-    fi.setFile( QString( SYSCONFDIR "/%1").arg(Theme::instance()->appName()), exclFile );
+    fi.setFile(QString(SYSCONFDIR "/" + Theme::instance()->appName()), exclFile);
     if ( ! fi.exists() ) {
         // Prefer to return the preferred path! Only use the fallback location
         // if the other path does not exist and the fallback is valid.
@@ -487,8 +489,8 @@ QVariant ConfigFile::getValue(const QString& param, const QString& group,
         systemSetting = systemSettings.value(param, defaultValue);
     } else { // Windows
         QSettings systemSettings(QString::fromLatin1("HKEY_LOCAL_MACHINE\\Software\\%1\\%2")
-                                .arg(APPLICATION_VENDOR).arg(Theme::instance()->appName()),
-                                QSettings::NativeFormat);
+                                     .arg(APPLICATION_VENDOR, Theme::instance()->appName()),
+                                 QSettings::NativeFormat);
         if (!group.isEmpty()) {
             systemSettings.beginGroup(group);
         }
@@ -596,6 +598,16 @@ void ConfigFile::setNewBigFolderSizeLimit(bool isChecked, quint64 mbytes)
     setValue(useNewBigFolderSizeLimitC, isChecked);
 }
 
+bool ConfigFile::confirmExternalStorage() const
+{
+    return getValue(confirmExternalStorageC, QString(), true).toBool();
+}
+
+void ConfigFile::setConfirmExternalStorage(bool isChecked)
+{
+    setValue(confirmExternalStorageC, isChecked);
+}
+
 bool ConfigFile::promptDeleteFiles() const
 {
     QSettings settings(configFile(), QSettings::IniFormat);
@@ -611,7 +623,12 @@ void ConfigFile::setPromptDeleteFiles(bool promptDeleteFiles)
 bool ConfigFile::monoIcons() const
 {
     QSettings settings(configFile(), QSettings::IniFormat);
-    return settings.value(QLatin1String(monoIconsC), false).toBool();
+    bool monoDefault = false; // On Mac we want bw by default
+#ifdef Q_OS_MAC
+    // OEM themes are not obliged to ship mono icons
+    monoDefault = (0 == (strcmp("ownCloud",APPLICATION_NAME)));
+#endif
+    return settings.value(QLatin1String(monoIconsC), monoDefault).toBool();
 }
 
 void ConfigFile::setMonoIcons(bool useMonoIcons)
