@@ -3,7 +3,8 @@
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; version 2 of the License.
+ * the Free Software Foundation; either version 2 of the License, or
+ * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful, but
  * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
@@ -12,6 +13,7 @@
  */
 
 #include "excludedfiles.h"
+#include "utility.h"
 
 #include <QFileInfo>
 
@@ -45,6 +47,13 @@ void ExcludedFiles::addExcludeFilePath(const QString& path)
     _excludeFiles.insert(path);
 }
 
+#ifdef WITH_TESTING
+void ExcludedFiles::addExcludeExpr(const QString &expr)
+{
+    _csync_exclude_add(_excludesPtr, expr.toLatin1().constData());
+}
+#endif
+
 bool ExcludedFiles::reloadExcludes()
 {
     c_strlist_destroy(*_excludesPtr);
@@ -63,18 +72,27 @@ bool ExcludedFiles::isExcluded(
         const QString& basePath,
         bool excludeHidden) const
 {
-    if (!filePath.startsWith(basePath)) {
+    if (!filePath.startsWith(basePath, Utility::fsCasePreserving() ? Qt::CaseInsensitive : Qt::CaseSensitive)) {
         // Mark paths we're not responsible for as excluded...
         return true;
     }
 
-    QFileInfo fi(filePath);
     if( excludeHidden ) {
-        if( fi.isHidden() || fi.fileName().startsWith(QLatin1Char('.')) ) {
-            return true;
+        QString path = filePath;
+        // Check all path subcomponents, but to *not* check the base path:
+        // We do want to be able to sync with a hidden folder as the target.
+        while (path.size() > basePath.size()) {
+            QFileInfo fi(path);
+            if( fi.isHidden() || fi.fileName().startsWith(QLatin1Char('.')) ) {
+                return true;
+            }
+
+            // Get the parent path
+            path = fi.absolutePath();
         }
     }
 
+    QFileInfo fi(filePath);
     csync_ftw_type_e type = CSYNC_FTW_TYPE_FILE;
     if (fi.isDir()) {
         type = CSYNC_FTW_TYPE_DIR;
